package be.iminds.ilabt.jfed.lowlevel.api_wrapper;

import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.lowlevel.*;
import be.iminds.ilabt.jfed.lowlevel.api.AbstractGeniAggregateManager;
import be.iminds.ilabt.jfed.lowlevel.api.user_spec.UserSpec;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnection;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnectionProvider;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.util.GeniUrn;
import org.slf4j.LoggerFactory;

import javax.annotation.Nullable;
import java.net.URL;
import java.util.*;

/**
 * AggregateManagerWrapper
 * <p/>
 * Supports only geni_sfa v3 rspec!
 */
public abstract class AggregateManagerWrapper extends ApiWrapper {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(AggregateManagerWrapper.class);

    protected final Server amServer;
    //a bit of a hack, but the last reply of the AM is sometimes needed by code calling this.
    protected AbstractGeniAggregateManager.AggregateManagerReply lastAmReply = null;

    protected Map<String, ?> extraOptionsForCallsWithCredential;

    protected List<AnyCredential> speaksForCredentials;

    protected AggregateManagerWrapper(Logger logger, GeniUserProvider geniUserProvider,
                                      JFedConnectionProvider connectionProvider, JFedPreferences jFedPreferences,
                                      Server amServer) {
        super(logger, geniUserProvider, connectionProvider, jFedPreferences);

        assert jFedPreferences != null;

        this.amServer = amServer;
    }

    protected JFedConnection getConnection(ApiInfo.Api api) throws JFedException {
        GeniUser user = geniUserProvider.getLoggedInGeniUser();
//        Server userAuth = user.getUserAuthority();
        return connectionProvider.getConnectionByAuthority(user, amServer, api);
    }

    protected GeniUser getUser() {
        return geniUserProvider.getLoggedInGeniUser();
    }

    protected Server getUserAuthority() {
        GeniUser user = geniUserProvider.getLoggedInGeniUser();
        Server userAuth = user.getUserAuthorityServer();
        return userAuth;
    }

    public Server getAmServer() {
        return amServer;
    }

    @Override
    public void setSpeaksFor(List<AnyCredential> speaksForCredentials, @Nullable GeniUrn speakingForUrn) {
        super.setSpeaksFor(speaksForCredentials, speakingForUrn);

        //nothing to do here about speaksForUserUrn, override this method to support speaksFor in the AM API specific way!
        //    This is typically done using setExtraOptionsForCallsWithCredential
    }

    public void setExtraOptionsForCallsWithCredential(Map<String, ?> extraOptionsForCallsWithCredential) {
        this.extraOptionsForCallsWithCredential = extraOptionsForCallsWithCredential;
    }
    protected Hashtable addCredentialExtraOptionsToMap(Map<String, Object> extraOptions) {
        if (extraOptionsForCallsWithCredential == null) {
            if (extraOptions == null)
                return null;
            else
                return new Hashtable(extraOptions);
        }

        Hashtable res = new Hashtable();
        if (extraOptions != null) res.putAll(extraOptions);
        res.putAll(extraOptionsForCallsWithCredential);
        return res;
    }
    protected Hashtable addCredentialExtraOptions(Hashtable extraOptions) {
        if (extraOptionsForCallsWithCredential == null)
            return extraOptions;

        Hashtable res = new Hashtable();
        if (extraOptions != null) res.putAll(extraOptions);
        res.putAll(extraOptionsForCallsWithCredential);
        return res;
    }
    /**
     * @return raw XML-RPC hashtable result of GetVersion call.
     */
    public abstract Hashtable getVersion() throws JFedException;

    public String listResources(List<AnyCredential> userCredentials, boolean available) throws JFedException {
        if (amServer != null && amServer.getServerType() != null && amServer.getServerType().equalsIgnoreCase("FOAM"))
            return listResources(userCredentials, available, "GENI", "3"); //see FEDIBBTDEV-1392: workaround for bug at al2s (it should ignore case but doesn't)
        else
            return listResources(userCredentials, available, "geni", "3");
    }
    public abstract String listResources(List<AnyCredential> userCredentials, boolean available, String type, String version) throws JFedException;

    /**
     *  Creates a sliver by making an allocate and provision request
     */
    public abstract String createSliver(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, String rspec, List<UserSpec> userSpecs, Date expirationDate)
            throws JFedException;

    public abstract boolean hasSeperateAllocateAndProvision();
    public interface AllocateReply {
        String getManifest();
        List<GeniUrn> getSliverUrns();
    }
    public abstract AllocateReply allocate(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, String rspec, Date expirationDate)
            throws JFedException;
    public abstract AllocateReply allocate(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, String rspec, Date startDate, Date endDate)
            throws JFedException;

    public abstract String provisionAndStart(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, List<UserSpec> userSpecs, Date expirationDate)
            throws JFedException;


    //TODO this should be renamed to "deleteAllSliversInSlice" and a real deleteSliver should be added
    public abstract void deleteSliver(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials) throws JFedException;

    public abstract StatusDetails status(GeniUrn urn/*slice or sliver*/, List<AnyCredential> sliceCredentials) throws JFedException;

    public abstract String describe(GeniUrn urn/*slice or sliver*/, List<AnyCredential> sliceCredentials) throws JFedException;

    /**
     * Renews one or more slivers, or all slivers in a slice.
     *
     * If a slice URN is given, this will assume all slivers in that slice need to be renewed, and will only use the slice URN in the call if possible.
     * If slice URN is null, and sliver URNs is not null and not empty, this call will attempt to renew only these slivers.
     *
     * It is advised to pass all slivers URNs and the slice URN to this call when all slivers in a slice need to be renewed. But only slice URN should work.
     * It is not advised to pass only the sliver URNs if all slivers of the slice need to be renewed (AMv2 cannot do this and not all AMv3 handle this well if geni_single is set).
     *
     * @param sliceUrn the URN of the slice (may be null in some cases, but best not)
     * @param slivers the URN of the slivers to be renewed (may be null or empty)
     * @param sliceCredentials creadential for the slice
     * @param newExpire new expiration time
     * @throws JFedException
     */
    public abstract void renewSlivers(GeniUrn sliceUrn, List<GeniUrn> slivers, List<AnyCredential> sliceCredentials, Date newExpire) throws JFedException;

    public AbstractGeniAggregateManager.AggregateManagerReply getLastReply() {
        return lastAmReply;
    }

    public abstract boolean isExirationdateSetByCreateSliver();

    //special calls that not every API has. These might choose another implementation than the previous calls!
    public abstract boolean isShareLanSupported();

    public abstract void shareLan(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, String lanClientId, String newSharedLanName) throws JFedException;

    public abstract void unshareLan(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, String lanClientId, String sharedLanName) throws JFedException;

    public abstract boolean isRestartSupported();
    public abstract boolean isReloadOSSupported();
    public abstract boolean isOpenConsoleSupported();

    public abstract void restart(GeniUrn sliverUrn, List<AnyCredential> sliceCredentials) throws JFedException;
    public abstract void reloadOS(GeniUrn sliverUrn, List<AnyCredential> sliceCredentials) throws JFedException;
    public abstract URL openConsole(GeniUrn sliverUrn, List<AnyCredential> sliceCredentials) throws JFedException;

    public abstract boolean isEditSshKeysSupported();

    public abstract void editSshKeys(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, Collection<UserSpec> userSpecs) throws JFedException;

    public boolean areUserDiskImageFunctionsSupported() {
        return false;
    }

    public static interface ImageInfo {
        public String getUrn();
        public String getUrl();
    }
    public ImageInfo createImage(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, GeniUrn componentUrn, String imageName,
                                 boolean global, boolean updatePrepare) throws JFedException {
        throw new RuntimeException("Unsupported operation");
    }
    public List<ImageInfo> listImages(List<AnyCredential> userCredentials) throws JFedException {
        throw new RuntimeException("Unsupported operation");
    }
    public boolean deleteImage(List<AnyCredential> userCredentials, GeniUrn imageUrn) throws JFedException {
        throw new RuntimeException("Unsupported operation");
    }

    public static class NotEnoughFreeResourcesException extends JFedException {
        public NotEnoughFreeResourcesException(XMLRPCCallDetails xmlRpcResult, GeniResponseCode geniResponseCode) {
            super("not enough free resources", null, xmlRpcResult, geniResponseCode);
        }

        public NotEnoughFreeResourcesException(String message, XMLRPCCallDetails xmlRpcResult, GeniResponseCode geniResponseCode) {
            super(message, null, xmlRpcResult, geniResponseCode);
        }

        public NotEnoughFreeResourcesException(XmlRpcApiCallReply reply) {
            super("not enough free resources", reply);
        }
    }
    public static class VlanUnavailableException extends JFedException {
        //may be null
        private final Integer vlan;

        public VlanUnavailableException(Integer vlan, XMLRPCCallDetails xmlRpcResult, GeniResponseCode geniResponseCode) {
            super("vlan not available", null, xmlRpcResult, geniResponseCode);
            this.vlan = vlan;
        }

        public VlanUnavailableException(Integer vlan, String message, XMLRPCCallDetails xmlRpcResult, GeniResponseCode geniResponseCode) {
            super(message, null, xmlRpcResult, geniResponseCode);
            this.vlan = vlan;
        }

        public VlanUnavailableException(Integer vlan, XmlRpcApiCallReply reply) {
            super("vlan not available", reply);
            this.vlan = vlan;
        }

        public Integer getVlan() {
            return vlan;
        }
    }
    public static class AnyVlanUnsupportedException extends JFedException {
        public AnyVlanUnsupportedException(XMLRPCCallDetails xmlRpcResult, GeniResponseCode geniResponseCode) {
            super("vlan \"any\" not supported", null, xmlRpcResult, geniResponseCode);
        }

        public AnyVlanUnsupportedException(String message, XMLRPCCallDetails xmlRpcResult, GeniResponseCode geniResponseCode) {
            super(message, null, xmlRpcResult, geniResponseCode);
        }

        public AnyVlanUnsupportedException(XmlRpcApiCallReply reply) {
            super("vlan \"any\" not supported", reply);
        }
    }

    public static class NoAmException extends RuntimeException {
        public NoAmException() {
        }

        public NoAmException(String message) {
            super(message);
        }

        public NoAmException(String message, Throwable cause) {
            super(message, cause);
        }

        public NoAmException(Throwable cause) {
            super(cause);
        }

        public NoAmException(String message, Throwable cause, boolean enableSuppression, boolean writableStackTrace) {
            super(message, cause, enableSuppression, writableStackTrace);
        }
    }
}
